
## Copyright (c) 2012, Viktor Jonsson
## All rights reserved.

## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##     * Redistributions of source code must retain the above copyright
##       notice, this list of conditions and the following disclaimer.
##     * Redistributions in binary form must reproduce the above copyright
##       notice, this list of conditions and the following disclaimer in the
##       documentation and/or other materials provided with the distribution.
##     * Neither the names of University of Gothenburg and/or Chalmers, nor the
##       names of its contributors may be used to endorse or promote products
##       derived from this software without specific prior written permission.

## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
## ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
## WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
## DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
## DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
## (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
## LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
## ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
## SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import random
parser = argparse.ArgumentParser(description='Generate dna reads of choosen size from provided protein sequence')
parser.add_argument('-seq', help='the protein sequence',default='LVFMCAGPTSYWQNHEDKR')
parser.add_argument('-len', type=int, help='desired length of each read in residues',default=25)
parser.add_argument('-nr', type=int, help='number of reads to be generated',default=10)
parser.add_argument('-file',help='name of file for storing generated reads', default='reads.fna')
parser.add_argument('-nrC',type=int, help='number of different undeltying contigs', default=5)
args = parser.parse_args()
# print(args.seq)



def generate_contigs(sequence,nr):
    """Function generate_read
    
    Takes the given peptide sequence and generates the given number of wildtype
    DNA sequences by randomizing the codons. With a small probability part of the
    generated sequences will be copied from a common conserved sequences to 
    simulate overlapping contigs.
    

    input: 

        sequence - the peptide sequence in a string
        nr - the number of contigs to be generated
    
    Output:
        list containing the specified number of full length contigs

    Brief functionality:
    First generates a common conserved contig by iterating over the sequence.
    Continues generating random sequences but with a certain probability takes
    a stretch of codons from the conserved contig. All generated contigs inluding
    the conserved ones are included in the outputlist. 

    ToDo
    Include more ways of mixing contigs to make the downstream analysis more 
    realistic
    """

    start = 0
    end = len(sequence)
    conservedContig = []    
        
    
    
    for i in range(start,end):
        if sequence[i] == 'W' or sequence[i] == 'M':
            conservedContig.append(codonDictionary[sequence[i]])
        else:
            nrCodons = len(codonDictionary[sequence[i]])
            #print nrCodons #works so far
            codonNr = random.randint(0,nrCodons-1)
            #print(codonDictionary[sequence[i]][codonNr])
            conservedContig.append(codonDictionary[sequence[i]][codonNr])

    

    conservedProb = 0.01
    
    contigList = []
    for i in range(0,nr -1):
        tempList = []
        
        conserveMode = False
      
        
    
    
        for i in range(start,end):
    
            if random.random() < conservedProb:
                conserveMode = True
                counter = 0
            if conserveMode:
                
                
                tempList.append(conservedContig[i])
                counter += 1
                if counter > 20:

                    conserveMode = False

            else:
                if sequence[i] == 'W' or sequence[i] == 'M':
                    tempList.append(codonDictionary[sequence[i]])
                else:
                    nrCodons = len(codonDictionary[sequence[i]])
                    #print nrCodons #works so far
                    codonNr = random.randint(0,nrCodons-1)
                    #print(codonDictionary[sequence[i]][codonNr])
                    tempList.append(codonDictionary[sequence[i]][codonNr])

        contigList.append(''.join(tempList))
    
    contigList.append(''.join(conservedContig))
    return contigList
    
def generate_read(sequence,length,all):
    """Function generate_contigs
    
    Takes the given nucleotide sequence and extracts a random part of specified
    length

    input:

        sequence - the nucleotide sequence as a string
        length - the desired length of the final read in residue count
        all - boolean that states if the length of the desired reads are longer than 
            the total number of residues
    
    Output:
        String containing the generated read

    Brief functionality:
    Randomly generates a start and stop within the peptide sequence and extracts it

    ToDo
    Pad with nucleotides before and after for reads overlapping with the edges
    Implement the possibility of generating reveresd sequences
    Implement mutations
    """
    
    
    
    if all:
        start = 0
        end = len(sequence)
        
    else:
        flex = len(sequence)-length*3
        start = random.randint(0,flex-1)
        end = start + length*3
    
    
    #print start
    #print end

    return sequence[start:end]
    

if args.len > len(args.seq):
    all = True
else:
    all = False

#print args.len
#print len(args.seq)*3
#dictionary containing residue to codon map
codonDictionary = {'I': ('ATT', 'ATC', 'ATA') ,
'L': ('CTT', 'CTC', 'CTA', 'CTG', 'TTA', 'TTG') ,
'V': ('GTT', 'GTC', 'GTA', 'GTG') ,
'F': ('TTT', 'TTC') ,
'M': ('ATG') ,
'C': ('TGT', 'TGC') ,
'A': ('GCT', 'GCC', 'GCA', 'GCG') ,
'G': ('GGT', 'GGC', 'GGA', 'GGG') ,
'P': ('CCT', 'CCC', 'CCA', 'CCG') ,
'T': ('ACT', 'ACC', 'ACA', 'ACG') ,
'S': ('TCT', 'TCC', 'TCA', 'TCG', 'AGT', 'AGC') ,
'Y': ('TAT', 'TAC') ,
'W': ('TGG') ,
'Q': ('CAA', 'CAG') ,
'N': ('AAT', 'AAC') ,
'H': ('CAT', 'CAC') ,
'E': ('GAA', 'GAG') ,
'D': ('GAT', 'GAC') ,
'K': ('AAA', 'AAG') ,
'R': ('CGT', 'CGC', 'CGA', 'CGG', 'AGA', 'AGG') ,
'Stop': ('TAA', 'TAG', 'TGA')} 

#print(generate_read(args.seq,args.len,all))

#build the desired number of contigs
contigs = generate_contigs(args.seq,args.nrC)
#print contigs

#the main loop of the program:
#opens the file to write to and continously adds new reads to the file
f = open(args.file,'w')
for j in range(0,args.nr):
    
    f.write( '>visaw' + str(j) + '\n')
    
    f.write(generate_read(contigs[random.randint(0,args.nrC-1)],args.len,all) + '\n')
    f.write('\n')

f.close()

f = open('contigs_'+args.file,'w')
for j in range(0,args.nrC):
    
    f.write( '>visaw_contig' + str(j) + '\n')
    
    f.write(contigs[j] + '\n')
    f.write('\n')

f.close()


#print(codonDictionary.keys())
#print(codonDictionary.values())


